@@ -20,6 +20,7 @@
 #include "php_apc.h"
 #include "apc_iterator.h"
 #include "apc_cache.h"
+#include "apc_strings.h"
 
 #include "ext/standard/md5.h"
 #include "SAPI.h"
@@ -33,54 +34,64 @@ zend_class_entry* apc_iterator_get_ce(void) {
 }
 
 /* {{{ apc_iterator_item */
-static apc_iterator_item_t* apc_iterator_item_ctor(apc_iterator_t *iterator, apc_cache_slot_t *slot) {
-	zval zvalue;
-	apc_context_t ctxt = {0, };
+static apc_iterator_item_t* apc_iterator_item_ctor(
+		apc_iterator_t *iterator, apc_cache_entry_t *entry) {
+	zval zv;
+	HashTable *ht;
 	apc_iterator_item_t *item = ecalloc(1, sizeof(apc_iterator_item_t));
 
 	array_init(&item->value);
+	ht = Z_ARRVAL(item->value);
 
-	item->key = zend_string_dup(slot->key.str, 0);
+	item->key = zend_string_dup(entry->key, 0);
 
 	if (APC_ITER_TYPE & iterator->format) {
-		add_assoc_string_ex(&item->value, "type", sizeof("type")-1, "user");
+		ZVAL_STR_COPY(&zv, apc_str_user);
+		zend_hash_add_new(ht, apc_str_type, &zv);
 	}
 
 	if (APC_ITER_KEY & iterator->format) {
-		add_assoc_str(&item->value, "key", zend_string_copy(item->key));
+		ZVAL_STR_COPY(&zv, item->key);
+		zend_hash_add_new(ht, apc_str_key, &zv);
 	}
 
 	if (APC_ITER_VALUE & iterator->format) {
-		apc_cache_make_copy_out_context(apc_user_cache, &ctxt);
-		ZVAL_UNDEF(&zvalue);
-		apc_cache_fetch_zval(&ctxt, &zvalue, &slot->value->val);
-		add_assoc_zval(&item->value, "value", &zvalue);
-		apc_cache_destroy_context(&ctxt);
+		ZVAL_UNDEF(&zv);
+		apc_cache_entry_fetch_zval(apc_user_cache, entry, &zv);
+		zend_hash_add_new(ht, apc_str_value, &zv);
 	}
 
 	if (APC_ITER_NUM_HITS & iterator->format) {
-		add_assoc_long(&item->value, "num_hits", slot->nhits);
+		ZVAL_LONG(&zv, entry->nhits);
+		zend_hash_add_new(ht, apc_str_num_hits, &zv);
 	}
 	if (APC_ITER_MTIME & iterator->format) {
-		add_assoc_long(&item->value, "mtime", slot->key.mtime);
+		ZVAL_LONG(&zv, entry->mtime);
+		zend_hash_add_new(ht, apc_str_mtime, &zv);
 	}
 	if (APC_ITER_CTIME & iterator->format) {
-		add_assoc_long(&item->value, "creation_time", slot->ctime);
+		ZVAL_LONG(&zv, entry->ctime);
+		zend_hash_add_new(ht, apc_str_creation_time, &zv);
 	}
 	if (APC_ITER_DTIME & iterator->format) {
-		add_assoc_long(&item->value, "deletion_time", slot->dtime);
+		ZVAL_LONG(&zv, entry->dtime);
+		zend_hash_add_new(ht, apc_str_deletion_time, &zv);
 	}
 	if (APC_ITER_ATIME & iterator->format) {
-		add_assoc_long(&item->value, "access_time", slot->atime);
+		ZVAL_LONG(&zv, entry->atime);
+		zend_hash_add_new(ht, apc_str_access_time, &zv);
 	}
 	if (APC_ITER_REFCOUNT & iterator->format) {
-		add_assoc_long(&item->value, "ref_count", slot->value->ref_count);
+		ZVAL_LONG(&zv, entry->ref_count);
+		zend_hash_add_new(ht, apc_str_ref_count, &zv);
 	}
 	if (APC_ITER_MEM_SIZE & iterator->format) {
-		add_assoc_long(&item->value, "mem_size", slot->value->mem_size);
+		ZVAL_LONG(&zv, entry->mem_size);
+		zend_hash_add_new(ht, apc_str_mem_size, &zv);
 	}
 	if (APC_ITER_TTL & iterator->format) {
-		add_assoc_long(&item->value, "ttl", slot->value->ttl);
+		ZVAL_LONG(&zv, entry->ttl);
+		zend_hash_add_new(ht, apc_str_ttl, &zv);
 	}
 
 	return item;
@@ -157,21 +168,27 @@ zend_object* apc_iterator_create(zend_class_entry *ce) {
 /* {{{ apc_iterator_search_match
  *       Verify if the key matches our search parameters
  */
-static int apc_iterator_search_match(apc_iterator_t *iterator, apc_cache_slot_t *slot) {
+static int apc_iterator_search_match(apc_iterator_t *iterator, apc_cache_entry_t *entry) {
 	int rval = 1;
 
 #ifdef ITERATOR_PCRE
 	if (iterator->regex) {
 # if PHP_VERSION_ID >= 70300
-		rval = (pcre2_match(iterator->re, (PCRE2_SPTR) ZSTR_VAL(slot->key.str), ZSTR_LEN(slot->key.str), 0, 0, iterator->re_match_data, php_pcre_mctx()) >= 0);
+		rval = pcre2_match(
+			php_pcre_pce_re(iterator->pce),
+			(PCRE2_SPTR) ZSTR_VAL(entry->key), ZSTR_LEN(entry->key),
+			0, 0, iterator->re_match_data, php_pcre_mctx()) >= 0;
 # else
-		rval = (pcre_exec(iterator->re, NULL, ZSTR_VAL(slot->key.str), ZSTR_LEN(slot->key.str), 0, 0, NULL, 0) >= 0);
+		rval = pcre_exec(
+			iterator->pce->re, iterator->pce->extra,
+			ZSTR_VAL(entry->key), ZSTR_LEN(entry->key),
+			0, 0, NULL, 0) >= 0;
 # endif
 	}
 #endif
 
 	if (iterator->search_hash) {
-		rval = zend_hash_exists(iterator->search_hash, slot->key.str);
+		rval = zend_hash_exists(iterator->search_hash, entry->key);
 	}
 
 	return rval;
@@ -179,14 +196,10 @@ static int apc_iterator_search_match(apc_iterator_t *iterator, apc_cache_slot_t
 /* }}} */
 
 /* {{{ apc_iterator_check_expiry */
-static int apc_iterator_check_expiry(apc_cache_t* cache, apc_cache_slot_t *slot, time_t t)
+static int apc_iterator_check_expiry(apc_cache_t* cache, apc_cache_entry_t *entry, time_t t)
 {
-	if(slot->value->ttl) {
-		if((time_t) (slot->ctime + slot->value->ttl) < t) {
-			return 0;
-		}
-	} else if(cache->ttl) {
-		if(slot->ctime + cache->ttl < t) {
+	if (entry->ttl) {
+		if ((time_t) (entry->ctime + entry->ttl) < t) {
 			return 0;
 		}
 	}
@@ -197,12 +210,9 @@ static int apc_iterator_check_expiry(apc_cache_t* cache, apc_cache_slot_t *slot,
 
 /* {{{ apc_iterator_fetch_active */
 static int apc_iterator_fetch_active(apc_iterator_t *iterator) {
-	int count=0;
-	apc_cache_slot_t *slot;
+	int count = 0;
 	apc_iterator_item_t *item;
-	time_t t;
-
-	t = apc_time();
+	time_t t = apc_time();
 
 	while (apc_stack_size(iterator->stack) > 0) {
 		apc_iterator_item_dtor(apc_stack_pop(iterator->stack));
@@ -211,18 +221,18 @@ static int apc_iterator_fetch_active(apc_iterator_t *iterator) {
 	APC_RLOCK(apc_user_cache->header);
 	php_apc_try {
 		while (count <= iterator->chunk_size && iterator->slot_idx < apc_user_cache->nslots) {
-			slot = apc_user_cache->slots[iterator->slot_idx];
-			while (slot) {
-				if (apc_iterator_check_expiry(apc_user_cache, slot, t)) {
-					if (apc_iterator_search_match(iterator, slot)) {
+			apc_cache_entry_t *entry = apc_user_cache->slots[iterator->slot_idx];
+			while (entry) {
+				if (apc_iterator_check_expiry(apc_user_cache, entry, t)) {
+					if (apc_iterator_search_match(iterator, entry)) {
 						count++;
-						item = apc_iterator_item_ctor(iterator, slot);
+						item = apc_iterator_item_ctor(iterator, entry);
 						if (item) {
 							apc_stack_push(iterator->stack, item);
 						}
 					}
 				}
-				slot = slot->next;
+				entry = entry->next;
 			}
 			iterator->slot_idx++;
 		}
@@ -237,27 +247,26 @@ static int apc_iterator_fetch_active(apc_iterator_t *iterator) {
 
 /* {{{ apc_iterator_fetch_deleted */
 static int apc_iterator_fetch_deleted(apc_iterator_t *iterator) {
-	int count=0;
-	apc_cache_slot_t *slot;
+	int count = 0;
 	apc_iterator_item_t *item;
 
 	APC_RLOCK(apc_user_cache->header);
 	php_apc_try {
-		slot = apc_user_cache->header->gc;
-		while (slot && count <= iterator->slot_idx) {
+		apc_cache_entry_t *entry = apc_user_cache->header->gc;
+		while (entry && count <= iterator->slot_idx) {
 			count++;
-			slot = slot->next;
+			entry = entry->next;
 		}
 		count = 0;
-		while (slot && count < iterator->chunk_size) {
-			if (apc_iterator_search_match(iterator, slot)) {
+		while (entry && count < iterator->chunk_size) {
+			if (apc_iterator_search_match(iterator, entry)) {
 				count++;
-				item = apc_iterator_item_ctor(iterator, slot);
+				item = apc_iterator_item_ctor(iterator, entry);
 				if (item) {
 					apc_stack_push(iterator->stack, item);
 				}
 			}
-			slot = slot->next;
+			entry = entry->next;
 		}
 	} php_apc_finally {
 		iterator->slot_idx += count;
@@ -271,20 +280,22 @@ static int apc_iterator_fetch_deleted(apc_iterator_t *iterator) {
 
 /* {{{ apc_iterator_totals */
 static void apc_iterator_totals(apc_iterator_t *iterator) {
-	apc_cache_slot_t *slot;
+	time_t t = apc_time();
 	int i;
 
 	APC_RLOCK(apc_user_cache->header);
 	php_apc_try {
 		for (i=0; i < apc_user_cache->nslots; i++) {
-			slot = apc_user_cache->slots[i];
-			while (slot) {
-				if (apc_iterator_search_match(iterator, slot)) {
-					iterator->size += slot->value->mem_size;
-					iterator->hits += slot->nhits;
-					iterator->count++;
+			apc_cache_entry_t *entry = apc_user_cache->slots[i];
+			while (entry) {
+				if (apc_iterator_check_expiry(apc_user_cache, entry, t)) {
+					if (apc_iterator_search_match(iterator, entry)) {
+						iterator->size += entry->mem_size;
+						iterator->hits += entry->nhits;
+						iterator->count++;
+					}
 				}
-				slot = slot->next;
+				entry = entry->next;
 			}
 		}
 	} php_apc_finally {
@@ -334,16 +345,17 @@ void apc_iterator_obj_init(apc_iterator_t *iterator, zval *search, zend_long for
 	if (search && Z_TYPE_P(search) == IS_STRING && Z_STRLEN_P(search)) {
 #ifdef ITERATOR_PCRE
 		iterator->regex = zend_string_copy(Z_STR_P(search));
-		iterator->re = pcre_get_compiled_regex(iterator->regex, NULL, NULL);
+		iterator->pce = pcre_get_compiled_regex_cache(iterator->regex);
 
-		if (!iterator->re) {
+		if (!iterator->pce) {
 			apc_error("Could not compile regular expression: %s", Z_STRVAL_P(search));
 			zend_string_release(iterator->regex);
 			iterator->regex = NULL;
 		}
 
 # if PHP_VERSION_ID >= 70300
-		iterator->re_match_data = pcre2_match_data_create_from_pattern(iterator->re, php_pcre_gctx());
+		iterator->re_match_data = pcre2_match_data_create_from_pattern(
+			php_pcre_pce_re(iterator->pce), php_pcre_gctx());
 # endif
 #else
 		apc_error("Regular expressions support is not enabled, please enable PCRE for " APC_ITERATOR_NAME " regex support.");
@@ -555,7 +567,7 @@ ZEND_END_ARG_INFO()
 
 /* {{{ apc_iterator_functions */
 static zend_function_entry apc_iterator_functions[] = {
-	PHP_ME(apc_iterator, __construct, arginfo_apc_iterator___construct, ZEND_ACC_PUBLIC|ZEND_ACC_CTOR)
+	PHP_ME(apc_iterator, __construct, arginfo_apc_iterator___construct, ZEND_ACC_PUBLIC)
 	PHP_ME(apc_iterator, rewind, arginfo_apc_iterator_void, ZEND_ACC_PUBLIC)
 	PHP_ME(apc_iterator, current, arginfo_apc_iterator_void, ZEND_ACC_PUBLIC)
 	PHP_ME(apc_iterator, key, arginfo_apc_iterator_void, ZEND_ACC_PUBLIC)
@@ -603,6 +615,10 @@ int apc_iterator_init(int module_number) {
 }
 /* }}} */
 
+int apc_iterator_shutdown(int module_number) {
+	return SUCCESS;
+}
+
 /* {{{ apc_iterator_delete */
 int apc_iterator_delete(zval *zobj) {
 	apc_iterator_t *iterator;
